/*
 * File: genlib.h
 * Version: 1.1
 *
 * Changes have been noted by the marker [DET].
 * -----------------------------------------------------
 * This file contains several definitions that form the
 * core of a general-purpose ANSI C library developed by Eric
 * Roberts.  The goal of this library is to provide a basic
 * set of tools and conventions that increase the readability
 * of C programs, particularly as they are used in a teaching
 * environment.
 *
 * The basic definitions provided by genlib.h are:
 *
 *    1.  Declarations for several new "primitive" types
 *        (most importantly string) that are used throughout
 *        the other libraries and applications as fundamental
 *        types.
 *
 *    2.  A new set of functions for memory allocation.
 *
 *    3.  A function for error handling.
 *
 *    4.  A repeat statement for loops with interior exits.
 */

#ifndef _genlib_h
#define _genlib_h

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>

#include <string>               // [DET]
using namespace std;            // [DET]


/* Section 1 -- Define new "primitive" types */

/*
 * Type: bool
 * ----------
 * [DET] Original definition deleted. Use the C++ bool type instead.
 */

#undef TRUE
#undef FALSE
#define TRUE true
#define FALSE false

/*
 * Type: string
 * ------------
 * [DET] Original definition deleted. Use the C++ string type instead.
 */

/*
 * Type: stream
 * ------------
 * Like string, the stream type is used to provide additional
 * readability and is defined to be equivalent to FILE *
 * (which is particularly confusing because it violates
 * standard case conventions).  This type is not used in
 * the text but is preserved in genlib.h, so it is possible
 * to teach all of CS1 without exposing any pointers.
 */

typedef FILE *stream;

/*
 * Constant: UNDEFINED
 * -------------------
 * Besides NULL, the only other constant of pointer type is
 * UNDEFINED, which is used in certain packages as a special
 * sentinel to indicate an undefined pointer value.  In many
 * such contexts, NULL is a legitimate data value and is
 * therefore inappropriate as a sentinel.
 */

extern void const * UNDEFINED;


/* Section 2 -- Memory allocation */

/*
 * [DET] Original definitions deleted. Use the standard C++ new 
 * and delete operators instead.
 */

/*
 * Function: GetBlock
 * Usage: ptr = (type) GetBlock(nbytes);
 *
 * Instead use: ptr = new type;                 [DET]
 * -------------------------------------
 * GetBlock allocates a block of memory of the given size.  If
 * no memory is available, GetBlock generates an error.
 */

/*
 * Function: FreeBlock
 * Usage: FreeBlock(ptr);
 *
 * Instead use: delete ptr;                     [DET]
 * ----------------------
 * FreeBlock frees the memory associated with ptr, which must
 * have been allocated using GetBlock, New, or NewArray.
 */

/*
 * Macro: New
 * Usage: p = New(pointer-type);
 *
 * Instead use: p = new type;                   [DET]
 * -----------------------------
 * The New pseudofunction allocates enough space to hold an
 * object of the type to which pointer-type points and returns
 * a pointer to the newly allocated pointer.  Note that
 * "New" is different from the "new" operator used in C++;
 * the former takes a pointer type and the latter takes the
 * target type.
 */

/*
 * Macro: NewArray
 * Usage: p = NewArray(n, element-type);
 *
 * Instead use: p = new element-type[n];      [DET]
 * -------------------------------------
 * NewArray allocates enough space to hold an array of n
 * values of the specified element type.
 */


/* Section 3 -- Basic error handling */

/*
 * Function: Error
 * Usage: Error(msg, ...)
 * ----------------------
 * Error generates an error string, expanding % constructions
 * appearing in the error message string just as printf does.
 * The program exits with a status code indicating failure (as 
 * given by the constant ErrorExitStatus).  The length of the 
 * error message string following expansion must not exceed
 * MaxErrorMessage, and it is the client's responsibility
 * to ensure this.
 */

void Error(string msg, ...);


/* Section 4 -- The repeat pseudo-statement */

/*
 * Statement form: repeat { ... }
 * ------------------------------
 * Some instructors who have taught CS1 using this library
 * have found that using
 *
 *     while (TRUE)
 *
 * to initiate a loop with an interior exit is confusing to
 * students, particularly when it comes at the beginning of
 * the course.  This macro defines "repeat" as an infinite
 * loop construct for instructors who find it easier to
 * explain, although it is not used in the text.   Similar
 * macro definitions are common in industry.
 */

#define repeat for (;;)


/* Section 5 -- Special code for the Borland C compiler */

/*
 * [DET] Original section was obsolete and therefore deleted.
 */


#endif
