// maze.c
// Jace Miller

#include "maze.h"

const int MIN_SIZE = 1;
const int MAX_SIZE = 25;

const int UP        = 0;
const int RIGHT     = 1;
const int DOWN      = 2;
const int LEFT      = 3;

const int FORWARD       = 0;
const int TURN_LEFT     = 1;
const int TURN_RIGHT    = 2;

const double WALL_LENGTH		= 0.25;
const double MAZE_LEFT_OFFSET	= 0.22;
const double MAZE_TOP_OFFSET	= 6.45;

int direction;
int last_direction;
int theseus_x, theseus_y;
int maze_width;     // measured in cells
int maze_height;    // measured in cells
bool wallsHorizontal[MAX_SIZE + 1][MAX_SIZE + 1];
bool wallsVertical[MAX_SIZE + 1][MAX_SIZE + 1];
bool occupiedNodes[MAX_SIZE + 1][MAX_SIZE + 1];
int delay_in_ms = 0;

void TurnOnHorizontalWall(int x, int y);
void TurnOnVerticalWall(int x, int y);

void GenerateMaze(int width, int height, int delay)
{
    bool chooseHorizontal = false;
    bool emptyWallSpotsExist = true;
    int possibleWallCount = 0;
    int wallToTurnOn = 0;

    if(delay >= 0)
    {
        delay_in_ms = delay;
    }

    srand(time(NULL));

    // constrain maze dimensions
    if(width < MIN_SIZE)
    {
        width = MIN_SIZE;
    }
    if(height < MIN_SIZE)
    {
        height = MIN_SIZE;
    }
    if(width > MAX_SIZE)
    {
        width = MAX_SIZE;
    }
    if(height > MAX_SIZE)
    {
        height = MAX_SIZE;
    }
    maze_width = width;
    maze_height = height;

    // turn off all walls and clear all nodes
    for(int i = 0; i <= MAX_SIZE; ++i)
    {
        for(int j = 0; j <= MAX_SIZE; ++j)
        {
            if(i < MAX_SIZE && j < MAX_SIZE)
            {
                wallsHorizontal[i][j] = false;
                wallsVertical[i][j] = false;
            }
            occupiedNodes[i][j] = false;
        }
    }

    // turn on outermost wall boundary and mark edge nodes as occupied
    // top and bottom walls
    for(int i = 0; i < maze_width; ++i)
    {
        TurnOnHorizontalWall(i, 0);
        TurnOnHorizontalWall(i, maze_height);
    }
    // left and right walls
    for(int i = 0; i < maze_height; ++i)
    {
        TurnOnVerticalWall(0, i);
        TurnOnVerticalWall(maze_width, i);
    }

    // draw wall from an occupied node to an unoccupied node
    emptyWallSpotsExist = true;
    while(true == emptyWallSpotsExist)
    {
        // randomly choose vert or horiz wall
        possibleWallCount = 0;
        chooseHorizontal = false;
        if(rand() < (RAND_MAX / 2))
        {
            chooseHorizontal = true;
        }

        if(true == chooseHorizontal)
        {
            // count number of possible horizontal walls
            for(int i = 0; i < maze_width; ++i)
            {
                for(int j = 0; j < maze_height; ++j)
                {
                    // if exactly one of the nodes of this wall is occupied
                    if( (occupiedNodes[i][j] && !occupiedNodes[i+1][j]) ||
                        (!occupiedNodes[i][j] && occupiedNodes[i+1][j]) )
                    {
                        ++possibleWallCount;
                    }
                }
            }
        }
        // if there were no more possible spots to draw horizontal walls,
        // then try looking for possible spots to draw vertical walls
        if(0 == possibleWallCount)
        {
            chooseHorizontal = false;
        }
        if(false == chooseHorizontal)
        {
            // count number of possible vertical walls
            for(int i = 0; i < maze_width; ++i)
            {
                for(int j = 0; j < maze_height; ++j)
                {
                    // if exactly one of the nodes of this wall is occupied
                    if( (occupiedNodes[i][j] && !occupiedNodes[i][j+1]) ||
                        (!occupiedNodes[i][j] && occupiedNodes[i][j+1]) )
                    {
                        ++possibleWallCount;
                    }
                }
            }
        }
        // if there are no horizontal or vertical walls, then
        // no more walls can be added and the maze is complete!
        if(0 == possibleWallCount)
        {
            emptyWallSpotsExist = false;
            break;
        }
        // random number between 0 and number of possible walls
        wallToTurnOn = rand() * possibleWallCount / RAND_MAX;

        // go through walls and turn on the (random number)th wall
        if(true == chooseHorizontal)
        {
            for(int i = 0; i < maze_width; ++i)
            {
                for(int j = 0; j < maze_height; ++j)
                {
                    // only look at possible places
                    if( (occupiedNodes[i][j] && !occupiedNodes[i+1][j]) ||
                        (!occupiedNodes[i][j] && occupiedNodes[i+1][j]) )
                    {
                        if(0 == wallToTurnOn)
                        {
                            TurnOnHorizontalWall(i, j);
                        }
                        --wallToTurnOn;
                    }
                }
            }
        }
        else
        {
            for(int i = 0; i < maze_width; ++i)
            {
                for(int j = 0; j < maze_height; ++j)
                {
                    // only look at possible places
                    if( (occupiedNodes[i][j] && !occupiedNodes[i][j+1]) ||
                        (!occupiedNodes[i][j] && occupiedNodes[i][j+1]) )
                    {
                        if(0 == wallToTurnOn)
                        {
                            TurnOnVerticalWall(i, j);
                        }
                        --wallToTurnOn;
                    }
                }
            }
        }
    }
    // It is possible to have some places where horizontal walls could be drawn,
    // but were not and the maze still stop generating

    // remove upper-left-most horizontal wall as exit
    wallsHorizontal[0][0] = false;

    // start Theseus in lower right corner, facing up
    theseus_x = rand() * (maze_width - 1) / RAND_MAX;
    theseus_y = rand() * (maze_height - 1) / RAND_MAX;
    direction = UP;
    last_direction = UP;
}

void TurnOnHorizontalWall(int x, int y)
{
    wallsHorizontal[x][y] = true;
    occupiedNodes[x][y] = true;
    occupiedNodes[x+1][y] = true;
}

void TurnOnVerticalWall(int x, int y)
{
    wallsVertical[x][y] = true;
    occupiedNodes[x][y] = true;
    occupiedNodes[x][y+1] = true;
}

void PrintMaze()
{
    InitGraphics("Maze Escape");
	SetColor(0.0, 0.0, 0.0);
    for(int i = 0; i <= maze_width; ++i)
    {
        for(int j = 0; j <= maze_height; ++j)
        {
            if(wallsHorizontal[i][j])
            {
                MovePen(MAZE_LEFT_OFFSET + i * WALL_LENGTH, MAZE_TOP_OFFSET + j * -WALL_LENGTH);
                DrawLine(WALL_LENGTH, 0.0);
            }
            if(wallsVertical[i][j])
            {
                MovePen(MAZE_LEFT_OFFSET + i * WALL_LENGTH, MAZE_TOP_OFFSET + j * -WALL_LENGTH);
                DrawLine(0.0, -WALL_LENGTH);
            }
        }
    }
    // green square marks start
    SetColor(0.0, 0.7, 0.0);
    MovePen(MAZE_LEFT_OFFSET + (theseus_x + 0.55) * WALL_LENGTH, MAZE_TOP_OFFSET + (theseus_y + 0.55) * -WALL_LENGTH);
    DrawLine(0.3 * WALL_LENGTH, 0);
    DrawLine(0, -0.3 * WALL_LENGTH);
    DrawLine(-0.3 * WALL_LENGTH, 0);
    DrawLine(0, 0.3 * WALL_LENGTH);
    DrawLine(0.3 * WALL_LENGTH, -0.3 * WALL_LENGTH);
    DrawLine(-0.3 * WALL_LENGTH, 0);
    DrawLine(0.3 * WALL_LENGTH, 0.3 * WALL_LENGTH);

	SetColor(0.0, 0.5, 1.0);
}

void MoveForward()
{
    if(IfFacingWall())
    {
        // cannot move forward through a wall
        return;
    }
    Sleep(delay_in_ms);
    switch(direction)
    {
        case UP:
			MovePen(MAZE_LEFT_OFFSET + (theseus_x + 0.75) * WALL_LENGTH, MAZE_TOP_OFFSET + (theseus_y + 0.5) * -WALL_LENGTH);

		    // blank, short dash, blank
            SetColor(1, 1, 1);
            DrawLine(0.0, 0.25 * WALL_LENGTH);
            SetColor(0, 0.5, 1.0);
            DrawLine(0.0, 0.5 * WALL_LENGTH);
            SetColor(1, 1, 1);
            DrawLine(0.0, 0.25 * WALL_LENGTH);

            // long straight line up
            if(UP == last_direction)
            {
                SetColor(0, 0.5, 1.0);
                DrawLine(0.0, -1.25 * WALL_LENGTH);
                DrawLine(0.0, 1.25 * WALL_LENGTH);
            }

            // erase small
            if(LEFT == last_direction)
            {
                SetColor(1, 1, 1);
                DrawLine(-0.25 * WALL_LENGTH, -0.75 * WALL_LENGTH);
                DrawLine(0.25 * WALL_LENGTH, 0.0);
            }

            // big-L right, then up
            if(RIGHT == last_direction)
            {
                SetColor(1, 1, 1);
                DrawLine(-0.5 * WALL_LENGTH, -1.25 * WALL_LENGTH);
                SetColor(0, 0.5, 1.0);
                DrawLine(0.5 * WALL_LENGTH, 0.0);
                DrawLine(0.0, 0.5 * WALL_LENGTH);
            }

            // U-turn opening up
            if(DOWN == last_direction)
            {
                SetColor(1, 1, 1);
                DrawLine(-0.5 * WALL_LENGTH, -0.75 * WALL_LENGTH);
                SetColor(0, 0.5, 1.0);
                DrawLine(0.0, -0.5 * WALL_LENGTH);
                DrawLine(0.5 * WALL_LENGTH, 0.0);
                DrawLine(0.0, WALL_LENGTH);
                SetColor(1, 1, 1);
                DrawLine(0.0, 0.25 * WALL_LENGTH);
            }

            theseus_y = theseus_y - 1;
            last_direction = UP;
            break;
        case RIGHT:
			MovePen(MAZE_LEFT_OFFSET + (theseus_x + 0.5) * WALL_LENGTH, MAZE_TOP_OFFSET + (theseus_y + 0.75) * -WALL_LENGTH);

		    // blank, short dash, blank
            SetColor(1, 1, 1);
            DrawLine(0.25 * WALL_LENGTH, 0.0);
            SetColor(0, 0.5, 1.0);
            DrawLine(0.5 * WALL_LENGTH, 0.0);
            SetColor(1, 1, 1);
            DrawLine(0.25 * WALL_LENGTH, 0.0);

            // long straight line right
            if(RIGHT == last_direction)
            {
                SetColor(0, 0.5, 1.0);
                DrawLine(-1.25 * WALL_LENGTH, 0.0);
                DrawLine(1.25 * WALL_LENGTH, 0.0);
            }

            // erase small
            if(UP == last_direction)
            {
                SetColor(1, 1, 1);
                DrawLine(-0.75 * WALL_LENGTH, 0.25 * WALL_LENGTH);
                DrawLine(0.0, -0.25 * WALL_LENGTH);
            }

            // big-L down, then right
            if(DOWN == last_direction)
            {
                SetColor(1, 1, 1);
                DrawLine(-1.25 * WALL_LENGTH, 0.5 * WALL_LENGTH);
                SetColor(0, 0.5, 1.0);
                DrawLine(0.0, -0.5 * WALL_LENGTH);
                DrawLine(0.5 * WALL_LENGTH, 0.0);
            }

            // U-turn opening right
            if(LEFT == last_direction)
            {
                SetColor(1, 1, 1);
                DrawLine(-0.25 * WALL_LENGTH, 0.0);
                SetColor(0, 0.5, 1.0);
                DrawLine(-0.75 * WALL_LENGTH, 0.0);
                DrawLine(0.0, 0.5 * WALL_LENGTH);
                DrawLine(0.5 * WALL_LENGTH, 0.0);
            }

            theseus_x = theseus_x + 1;
            last_direction = RIGHT;
            break;
        case DOWN:
			MovePen(MAZE_LEFT_OFFSET + (theseus_x + 0.25) * WALL_LENGTH, MAZE_TOP_OFFSET + (theseus_y + 0.5) * -WALL_LENGTH);

		    // blank, short dash, blank
            SetColor(1, 1, 1);
            DrawLine(0.0, -0.25 * WALL_LENGTH);
            SetColor(0, 0.5, 1.0);
            DrawLine(0.0, -0.5 * WALL_LENGTH);
            SetColor(1, 1, 1);
            DrawLine(0.0, -0.25 * WALL_LENGTH);

            // long straight line down
            if(DOWN == last_direction)
            {
                SetColor(0, 0.5, 1.0);
                DrawLine(0.0, 1.25 * WALL_LENGTH);
                DrawLine(0.0, -1.25 * WALL_LENGTH);
            }

            // erase small
            if(RIGHT == last_direction)
            {
                SetColor(1, 1, 1);
                DrawLine(0.25 * WALL_LENGTH, 0.75 * WALL_LENGTH);
                DrawLine(-0.25 * WALL_LENGTH, 0.0);
            }

            // big-L left, then down
            if(LEFT == last_direction)
            {
                SetColor(1, 1, 1);
                DrawLine(0.5 * WALL_LENGTH, 1.25 * WALL_LENGTH);
                SetColor(0, 0.5, 1.0);
                DrawLine(-0.5 * WALL_LENGTH, 0.0);
                DrawLine(0.0, -0.5 * WALL_LENGTH);
            }

            // U-turn opening down
            if(UP == last_direction)
            {
                SetColor(1, 1, 1);
                DrawLine(0.5 * WALL_LENGTH, 0.75 * WALL_LENGTH);
                SetColor(0, 0.5, 1.0);
                DrawLine(0.0, 0.5 * WALL_LENGTH);
                DrawLine(-0.5 * WALL_LENGTH, 0.0);
                DrawLine(0.0, -WALL_LENGTH);
                SetColor(1, 1, 1);
                DrawLine(0.0, -0.25 * WALL_LENGTH);
            }

            theseus_y = theseus_y + 1;
            last_direction = DOWN;
            break;
        case LEFT:
            MovePen(MAZE_LEFT_OFFSET + (theseus_x + 0.5) * WALL_LENGTH, MAZE_TOP_OFFSET + (theseus_y + 0.25) * -WALL_LENGTH);

		    // blank, short dash, blank
            SetColor(1, 1, 1);
            DrawLine(-0.25 * WALL_LENGTH, 0.0);
            SetColor(0, 0.5, 1.0);
            DrawLine(-0.5 * WALL_LENGTH, 0.0);
            SetColor(1, 1, 1);
            DrawLine(-0.25 * WALL_LENGTH, 0.0);

            // long straight line left
            if(LEFT == last_direction)
            {
                SetColor(0, 0.5, 1.0);
                DrawLine(1.25 * WALL_LENGTH, 0.0);
                DrawLine(-1.25 * WALL_LENGTH, 0.0);
            }

            // erase small
            if(DOWN == last_direction)
            {
                SetColor(1, 1, 1);
                DrawLine(0.75 * WALL_LENGTH, -0.25 * WALL_LENGTH);
                DrawLine(0.0, 0.25 * WALL_LENGTH);
            }

            // big-L up, then left
            if(UP == last_direction)
            {
                SetColor(1, 1, 1);
                DrawLine(1.25 * WALL_LENGTH, -0.5 * WALL_LENGTH);
                SetColor(0, 0.5, 1.0);
                DrawLine(0.0, 0.5 * WALL_LENGTH);
                DrawLine(-0.5 * WALL_LENGTH, 0.0);
            }

            // U-turn opening left
            if(RIGHT == last_direction)
            {
                SetColor(1,1,1);
                DrawLine(0.25 * WALL_LENGTH, 0.0);
                SetColor(0, 0.5, 1.0);
                DrawLine(0.75 * WALL_LENGTH, 0.0);
                DrawLine(0.0, -0.5 * WALL_LENGTH);
                DrawLine(-0.5 * WALL_LENGTH, 0.0);
            }

            theseus_x = theseus_x - 1;
            last_direction = LEFT;
            break;
    }
}

void TurnRight()
{
    ++direction;
    direction %= 4;
}

void TurnLeft()
{
    direction += 3;
    direction %= 4;
}

bool IfFacingWall()
{
    switch(direction)
    {
        case UP:
            return wallsHorizontal[theseus_x][theseus_y];
            break;
        case RIGHT:
            return wallsVertical[theseus_x + 1][theseus_y];
            break;
        case DOWN:
            return wallsHorizontal[theseus_x][theseus_y + 1];
            break;
        case LEFT:
            return wallsVertical[theseus_x][theseus_y];
            break;
    }
    return false;
}

bool IfOutside()
{
    if(theseus_x < 0 || theseus_y < 0 || theseus_x > maze_width || theseus_y > maze_height)
    {
        return true;
    }
    return false;
}
